#include <stdlib.h>
#include <string.h>
#include "queue.h"

#define CNOS 7

class PacketGPDRR;
class GPDRR;


class PacketGPDRR : public PacketQueue {
  PacketGPDRR(): pkts(0),prev(0),next(0),deficitCounter(0),turn(0),bcount(0),src(-1) {
    for (int i=0; i<CNOS; i++)
      {
	//pktcnt[i]=0;
	bytes[i]=0;
      }
  }
  friend GPDRR;
  protected :
  int pkts;
  //  int pktcnt[CNOS];
  int bytes[CNOS];
  int bcount;
  int src;    //to detect collisions keep track of actual src address
  PacketGPDRR *prev;
  PacketGPDRR *next;
  int deficitCounter; 
  int turn;
  inline PacketGPDRR * activate(PacketGPDRR *head) {
    if (head) {
      this->prev = head->prev;
      this->next = head;
      head->prev->next = this;
      head->prev = this;
      return head;
    }
    this->prev = this;
    this->next = this;
    return this;
  }
  inline PacketGPDRR * idle(PacketGPDRR *head) {
    if (head == this) {
      if (this->next == this)
	return 0;
      this->next->prev = this->prev;
      this->prev->next = this->next;
      return this->next;
    }
    this->next->prev = this->prev;
    this->prev->next = this->next;
    return head;
  }
  // remove the packet with flwid i in the pktQ
  Packet* PacketGPDRR::removei (int i,int off_ip_) 
  {  
    Packet *p;
    int done=0,k=0;
    p=lookup(0);
    while (p && !done) {
      p=lookup(k++);
      hdr_ip *iph =(hdr_ip*)p->access(off_ip_);
      if (iph->flowid() == i) done=1;
    }
    this->remove(p);
    return p;
  }
};


class GPDRR : public Queue {
  public :
  GPDRR();
  /*virtual int command(int argc, const char*const* argv);*/
  Packet *deque(void);
  void enque(Packet *pkt);
  int hash(Packet *pkt);
protected:
  int buckets_ ; //total number of flows allowed
  int blimit_;    //total number of bytes allowed across all flows
  int quantum_;  //total number of bytes that a flow can send
  int mask_;     /*if set hashes on just the node address otherwise on 
		   node+port address*/
  int thresh_;   // thresholds for output drop mechanism
  int bytecnt ; //cumulative sum of bytes across all flows
  int flwcnt ; //total number of active flows
  int levels[CNOS]; // per level number of packets
  PacketGPDRR *curr; //current active flow
  PacketGPDRR *gpdrr ; //pointer to the entire pdrr struct
  int off_ip_;
  inline PacketGPDRR *getMaxFlow (PacketGPDRR *curr,int level) { //returns flow with max pkts
    int i;
    PacketGPDRR *tmp;
    PacketGPDRR *maxflow=curr;
    for (i=0,tmp=curr; i < flwcnt; i++,tmp=tmp->next) {
      if (maxflow->bytes[level] < tmp->bytes[level])
	maxflow=tmp;
    }
    return maxflow;
  }
};

static class GPDRRClass : public TclClass {
public:
  GPDRRClass() : TclClass("Queue/GPDRR") {}
  TclObject* create(int argc, const char*const* argv) {
    return (new GPDRR);
  }
} class_gpdrr;

GPDRR::GPDRR()
{
  int i;
  buckets_=16;
  quantum_=250;
  gpdrr=0;
  curr=0;
  flwcnt=0;
  bytecnt=0;
  mask_=0;
  for (i=0;i<CNOS;i++)
    levels[i]=0;
  bind("buckets_",&buckets_);
  bind("blimit_",&blimit_);
  bind("quantum_",&quantum_);
  bind("mask_",&mask_);
  bind ("thresh_",&thresh_);
  bind ("off_ip_",&off_ip_);
}
 
void GPDRR::enque(Packet* pkt)
{
  PacketGPDRR *q,*remq;
  Packet *p;
  int which,i;

  hdr_cmn *ch=(hdr_cmn*)pkt->access(off_cmn_);
  hdr_ip *iph = (hdr_ip*)pkt->access(off_ip_);
  if (!gpdrr)
    gpdrr=new PacketGPDRR[buckets_];
  which= hash(pkt) % buckets_;
  q=&gpdrr[which];

  /*detect collisions here */
  int compare = (!mask_ ? ((int)iph->src()) : ((int)iph->src()&0xff00));
  if (q->src ==-1)
    q->src= compare;
  else
    if (q->src != compare)
      fprintf(stderr,"Collisions between %d and %d src addresses\n",q->src,(int)iph->src());      

  q->enque(pkt);
  q->bcount += ch->size();
  q->bytes[iph->flowid()] += ch->size();
  ++q->pkts;
  ++levels[iph->flowid()];
  //  ++q->pktcnt[iph->flowid()];
  bytecnt +=ch->size();

  if (q->pkts==1)
    {
      curr = q->activate(curr);
      q->deficitCounter=0;
      ++flwcnt;
    }
  while (bytecnt > blimit_) {
    Packet *p;
    hdr_cmn *remch;
    hdr_ip *remiph;

    /*get the least important level and remove the packet from the flow having the
      most of such packets */

    
    for (i=CNOS-1;i>=0;i--)
      if (levels[i]) break;

    remq=getMaxFlow(curr,i);
    p=remq->removei(i,off_ip_);

    remch=(hdr_cmn*)p->access(off_cmn_);
    remiph=(hdr_ip*)p->access(off_ip_);
    bytecnt -= remch->size();
    levels[remiph->flowid()]--;
    if (remiph->flowid()==0)
      fprintf (stderr,"dropping the most imp pkt due to lack of buffer space\n");
    //    --remq->pktcnt[i];
    remq->bytes[i] -= remch->size();
    remq->bcount -= remch->size();
    --remq->pkts;
    if (remq->pkts==0) {
      curr=remq->idle(curr);
      --flwcnt;
    }
    drop(p);
  }
}

Packet *GPDRR::deque(void) 
{
  hdr_ip *iph;
  hdr_cmn *ch;
  int sum,k,dlevel;
  Packet *pkt=0;
  int done=0;
  
  while (!pkt) {
    if (bytecnt==0) {
      //fprintf (stderr,"No active flow\n");
      return(0);
    }
    


    if (!curr->turn) {
      curr->deficitCounter+=quantum_;
      curr->turn=1;
    }

    /* drop the pkts till threshold exceeded */
    while (!done ) {
      pkt = curr->lookup(0);
      iph=(hdr_ip*)pkt->access(off_ip_);
      ch=(hdr_cmn*)pkt->access(off_cmn_);
      { 
	sum=0;
	dlevel=iph->flowid();
	/* implement global priority so the  sum of all pkts of more imp levels */
	for (k=dlevel-1; k >=0; k--) 
	  sum += levels[k];
	    
	if (sum > thresh_) {
	  pkt=curr->deque();
	  iph=(hdr_ip*)pkt->access(off_ip_);
	  ch=(hdr_cmn*)pkt->access(off_cmn_);
	  curr->bcount -= ch->size();
	  --curr->pkts;
	  //	  --curr->pktcnt[dlevel];

	  //if (curr->bytes[dlevel] == ch->size()) 
	  //  printf ("Dropping the last packet in this level");
	  curr->bytes[dlevel] -= ch->size();
	  levels[dlevel]--;
	  bytecnt -= ch->size();
	  if (curr->pkts== 0) {
	    curr->turn=0;
	    --flwcnt;
	    curr->deficitCounter=0;
	    //	    if (flwcnt != 0 ) {
	    curr=curr->idle(curr);
	    curr->deficitCounter+=quantum_;
	    curr->turn=1;
	      //	    }
	  }
	  drop(pkt);
	}
      else
	done =1;
      }
    }

    if ( curr->deficitCounter >= ch->size()) {
      curr->deficitCounter -= ch->size();
      pkt=curr->deque();
      iph=(hdr_ip*)pkt->access(off_ip_);
      ch=(hdr_cmn*)pkt->access(off_cmn_);
      --curr->pkts;
      curr->bcount -= ch->size();
      //      --curr->pktcnt[iph->flowid()];
      curr->bytes[iph->flowid()] -= ch->size();
      bytecnt -= ch->size();
      levels[iph->flowid()]--;
      if (curr->pkts == 0) {
	curr->turn=0;
	--flwcnt;
	curr->deficitCounter=0;
	curr=curr->idle(curr);
      }
      return pkt;
    }
    else {
      curr->turn=0;
      curr=curr->next;
      pkt=0;
    }
  }
}

int GPDRR::hash(Packet* pkt)
{
  hdr_ip *iph=(hdr_ip*)pkt->access(off_ip_);
  int i;
  if (mask_)
     i = (int)iph->src() & (0xff00);
  else
    i = (int)iph->src();
  return ((i + (i >> 8) + ~(i>>4)) % ((2<<23)-1))+1; // modulo a large prime
  /*return i;*/
}
