#include "layer.h"
#include "rtp.h"
#include "Tcl.h"
#include "packet.h"
#include "random.h"
#include "ip.h"

static class LayerClass : public TclClass {
  public :
  LayerClass() : TclClass ("Agent/Layer") {}
  TclObject* create(int argc, const char*const* argv) {
    return (new Layer_Agent());
  }
} class_Layer;

void LayerTimer::expire(Event *)
{
  l_->timeout();
}

Layer_Agent::Layer_Agent() : Agent(PT_CBR),timer_(this),lasttime_(0),running_(0),seqno_(-1)
{
  bind ("packetSize_", &size_);
  bind("noofpkts_",&noofpkts_);
  bind_time("interval_",&interval_);
  bind("off_rtp_",&off_rtp_);
}


void Layer_Agent::start()
{
  double t;
  running_ = 1;
  vtime_ =0;
  //schedule a timer at sometime between 0 and interval_
  t = Random::uniform(0,interval_);
  lasttime_ = t;
  timer_.sched(t);
  vtime_ += interval_;
}



void Layer_Agent::stop()
{
  //cancel(0);
  running_=0;
}

void Layer_Agent::timeout ()
{
  double t;
  if (running_) {
      send_all_pkts();
      //schedule another timer between vtime_ and vtime + interval_
      t=Random::uniform(vtime_,vtime_+interval_);
      timer_.resched(t-lasttime_);
      lasttime_ = t;
      vtime_ += interval_;
  }
}

void Layer_Agent::send_all_pkts()
{
  int i;
  if (fid_ == 1) {
    //    printf ("Sending out packets now for layer 1 %g\n",Scheduler::instance().clock());
  }
  for (i=0; i<noofpkts_;i++)
    {
      Packet *p =allocpkt();
      hdr_ip *iph = (hdr_ip*)p->access(off_ip_);
      if (iph->flowid() != fid_)
	printf ("Bug here : %d %d\n",iph->flowid(),fid_);
      hdr_rtp* rh = (hdr_rtp*)p->access(off_rtp_);
      rh->seqno() = ++seqno_;
      target_->recv(p);
    }
}

/*
 * $layer start
 * $layer stop
 */
int Layer_Agent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "start") == 0) {
			start();
			return (TCL_OK);
		}
		if (strcmp(argv[1], "stop") == 0) {
			stop();
			return (TCL_OK);
		}
	}
	return (Agent::command(argc, argv));
}
