Class CustomRecv


CustomRecv instproc addr {} {
	$self instvar addr
	return $addr
}

CustomRecv instproc node {} {
	$self instvar node
	return $node
}

CustomRecv instproc subscription {} {
	$self instvar subscription
	return $subscription
}

CustomRecv instproc dumpLevel {} {
    global TraceFile rates
    if [info exists TraceFile] {
	$self instvar subscription node rateMap
	#XXX
	if ![info exists rateMap] {
	    set s 0
	    set rateMap "0"
	    foreach r $rates {
		set s [expr $s + $r]
		lappend rateMap $s
	    }
	}
	set r [lindex $rateMap $subscription]
	puts $TraceFile "[$node id] [ns-now] $r"
    }
}

CustomRecv instproc drop-layer {} {
    $self dumpLevel
    $self instvar subscription layer
    set n $subscription
    #
    # if we have an active layer, drop it
    #
    if { $n > 0 } {
	$self debug "DRP-LAYER $n"
	rect_del $self $n
	$layer($n) leave-group 
	incr n -1
	set subscription $n
    }
    $self dumpLevel
}

CustomRecv instproc add-layer {} {
    $self dumpLevel
    $self instvar maxlevel subscription layer
    set n $subscription
    if { $n < $maxlevel } {
	$self debug "ADD-LAYER"
	incr n
	set subscription $n
	$layer($n) join-group
	rect_add $self $n
    }
    $self dumpLevel
}

proc recv_debug msg {
	global recv_debug_flag
	if $recv_debug_flag {
		puts stderr "[format %.05f [ns-now]] $msg"
	}
}

#XXX should have debug flag on each obj
CustomRecv instproc debug msg {
	$self instvar addr subscription node
	recv_debug "nd [$node id] $addr layer $subscription $msg"
}



CustomRecv instproc init { ns localNode baseGroup n } {
        $self next
        $self instvar node addr maxlevel ctrl\
	    total_bits subscription layer
	set node $localNode
	set addr $baseGroup
        set maxlevel $n

	set ctrl [$ns create-agent $node Agent/Message 0]
	$ctrl proc handle msg "recv_ctrl $self \$msg"
	#XXX
	set caddr [expr $baseGroup + 0x1000]
	$node join-group $ctrl $caddr
	$ctrl set dst $caddr

	#
	# we number the subscription level starting at 1.
	# level 0 means no groups are subscribed to.
	# 
	set i 1
	while { $i <= $n } {
		set layerNo [expr $i - 1]
		set dst [expr $baseGroup + $layerNo]
		set layer($i) [new Layer $ns $self $dst $layerNo MyLossTrace normal]
                $self instvar layers
                lappend layers $layer($i)
		incr i

	}

	set total_bits 0

	#
	# set the subscription level to 0 and call add_layer
	# to start out with at least one group
	#
	set subscription 0
    set i 1
    while { $i <= $n } {
        $self add-layer
	incr i
    }

}


CustomRecv instproc loss-stats { ns interval addr } {
    $ns at [expr [$ns now] + $interval] "$self loss-stats $ns $interval $addr"
    #puts stderr "*****At [$ns now]"
    $self get-all-layer-stats [$ns now] $addr
}

#gets loss stats from all layers at interval
CustomRecv instproc get-all-layer-stats { time addr } {
    global plots
    $self instvar layers
    set i 0
    foreach s $layers {
	$s instvar util
	set monitor [$s mon]
	set frac [$util trivial-loss-util [$monitor compute-recv-fraction]]
	$monitor instvar sumutil sumtime
	incr sumtime
	set sumutil [expr $sumutil + $frac ]
#	puts stderr "% pkts rcvd for layer $i : $frac"
	if { $time > 599 } {
	    puts $plots "$addr $i $time $sumtime $frac $sumutil"
	    flush $plots
	}
	incr i 
    }
}




proc build_custom_receiver { ns node baseGroup n } {

	set recvr [new CustomRecv $ns $node $baseGroup $n]

	global allrecvr
	lappend allrecvr $recvr

	return $recvr
}



Class MyLossTrace -superclass Agent/LossMonitor
#
#compute loss percentages for flow i
#as (#pkts recvd) * 100 /(#pkts rcvd + #pkts lost)
#
MyLossTrace instproc init {} {
    $self next
    $self instvar lastlost lastpkts sumutil sumtime mynlost data
    set lastlost 0
    set lastpkts 0
    set sumutil 0
    set sumtime 0
    set mynlost 0
    # to compute lag
#    set data [new Lag 10]
}


MyLossTrace instproc compute-recv-fraction {} {
    $self instvar mynlost_ npkts_ lastlost lastpkts sumutil sumtime

    if { $npkts_ != 0 }  {
	if { $npkts_ != $lastpkts } {
	    set frac [expr ($npkts_ - $lastpkts) * 1.0 / ($mynlost_ + $npkts_ - $lastlost - $lastpkts)]
	} else {
	    set frac 0
	}
    } else {
	set frac 0
    }
    
    set lastlost $mynlost_
    set lastpkts $npkts_

    return $frac
}


