#Necessary params
#<delay> <qType> <seed> <bandwidth> <sources> <scenario> <level> <dir> <traffic> <bgsrcType> <upinterval> <statinterval> <lsminterval> <srcType> <noofbuffers> <bglevel> <runno> <rlm_debug_flag>

#Example
#../../ns rlmsims.tcl delay=0.01 qType=DropTail seed=1 proto=rlm bandwidth=4e6 sources=8 runtime=600 scenario=1 level=7 dir=5-a traffic=5 bgsrcType=lsm upinterval=1.0 statinterval=1.0 lsminterval=1.0 srcType=lsm noofbuffers=60 bglevel=4 runno=0


set rectFile rect.out

set statinterval 2.0
set rcvTime 160.0
set packetSize 1000
set buffers 0.12
set burstylength 760
set lsminterval 1.0
set totalpkts [ open new.out w]
set rlm_debug_flag 0
set runtime 600.0

#Defaults to suppress tcl error messages
Agent/Layer set packetSize_ 0
Agent/Layer set noofpkts_ 0
Agent/Layer set interval_ 0

set seeds(1) 188312339
set seeds(2) 1072664641
set seeds(3) 694388766
set seeds(4) 2009044369
set seeds(5) 934100682
set seeds(6) 1972392646
set seeds(7) 1936856304
set seeds(8) 1598189534
set seeds(9) 1822174485
set seeds(10) 1871883252


Queue/DRR set buckets_ 100
Queue/DRR set mask_ 1
Queue/DRR set quantum_ $packetSize

Simulator set NumberInterfaces_ 1
 
Agent/LossMonitor set nlost_ 0
Agent/LossMonitor set npkts_ 0
Agent/LossMonitor set expected_ -1

set i 0
while { $i < $argc } {
	set L [split [lindex $argv $i] =]
	if { [llength $L] == 2 } {
		set var [lindex $L 0]
		set val [lindex $L 1]
		set $var $val
	}
	incr i
}


Queue/DropTail set limit_ [expr int($buffers*$bandwidth/(8*$packetSize))]
Queue/PriorityQueue set limit2_ [expr int($buffers*$bandwidth/(8*$packetSize))]
Queue/PriorityQueue set thresh_  [expr int([Queue/PriorityQueue set limit2_] *0.6)]

#queue/drop-tail set limit 15
Queue/DRR set blimit [expr int($buffers*$bandwidth/8)]



#if { $bgsrcType == "cbr" } {
#    set bgrates "32e3 64e3 128e3 256e3"
#} else {
    #set bgrates "vicdata/vic.2.trunc vicdata/vic.4.trunc vicdata/vic.8.trunc vicdata/vic.16.trunc"
#    set bgrates "6000e3"
#}


if [ info exists debug-link ] { 
    set debugfile [open "out.tr" w]
}

ns-random $seeds($seed)
$defaultRNG seed $seeds($seed)

set outfilename "$dir/$qType-rlm-$srcType-$bgsrcType-$noofbuffers-$runno-$delay"
set plots [open  $outfilename w]


Simulator instproc create-agent { node type pktClass } {

	$self instvar Agents PortID 
	set agent [new $type]
	$agent set class_ $pktClass
	$self attach-agent $node $agent

	$agent proc get var {
		return [$self set $var]
	}

	return $agent
}


Simulator instproc get-link { node1 node2 } {
    $self instvar link
    set id1 [$node1 id]
    set id2 [$node2 id]
    return $link($id1:$id2)
}


source rcvr.tcl

if { $level == 7 } {
    if { $srcType == "cbr" } {
	set rates "32e3 64e3 128e3 256e3 512e3 1024e3 2048e3"
    } else {
	set rates "vicdata/vic.2.trunc vicdata/vic.4.trunc vicdata/vic.8.trunc vicdata/vic.16.trunc vicdata/vic.32.trunc vicdata/vic.64.trunc vicdata/vic.64.trunc"
    }
} else {
    if { $level == 5 } {
	if { $srcType == "cbr" } {
	    set rates "32e3 64e3 128e3 256e3 512e3"
	} else {
	set rates "vicdata/vic.2 vicdata/vic.4 vicdata/vic.8 vicdata/vic.16 vicdata/vic.32"
	}
    } else {
	if { $level == 4 } {
	    if { $srcType == "cbr" } {
		set rates "32e3 64e3 128e3 256e3"
	    } else {
		set rates "vicdata/vic.2 vicdata/vic.4 vicdata/vic.8 vicdata/vic.16"
	    }
	} else {
	    if { $level == 3 } {
		if { $srcType == "cbr" } { 
		    set rates "32e3 64e3 128e3"
		} else {
		    set rates "vicdata/vic.2 vicdata/vic.4 vicdata/vic.8"
		}
	    } else {
		if { $level == 2 } {
		    if { $srcType == "cbr" } {
			set rates "32e3 64e3"
		    } else {
			set rates "vicdata/vic.2 vicdata/vic.4"
		    }
		} else {
		    if { $level == 1 } {
			if { $srcType == "cbr" } {
			    set rates "32e3"
			} else {
			    set rates "vicdata/vic.2"
			}
		    }
		}
	    }
	}
    }
}

#XXX nedd to remove these
set nlevel $level
#set filenames $rates

#rlm-init proc is defn in rlm-cmn.tcl
rlm_init $rectFile $nlevel $runtime  

Class Topology

Topology instproc init { simulator } {
    $self instvar addr_allocator ns
    set ns $simulator
    set addr_allocator [expr 0x8000]

    #XXXX
    proc ns-now {} "return \[$ns now]"
    global sim
    set sim $ns

}

Topology instproc mknode nn {
	$self instvar node ns
	if ![info exists node($nn)] {
		set node($nn) [$ns node]
	}
}


#
# build a link between nodes $a and $b
# if either node doesn't exist, create it as a side effect.
# (we don't build any sources)
#
Topology instproc build_link { a b delay args } {
    if { $a == $b } {
	puts stderr "link from $a to $b?"
	exit 1
	}
    #XXX
    global bandwidth qType buffers packetSize noofbuffers
    
    $self instvar node ns
    $self mknode $a
    $self mknode $b
    if { $args != "" } {
	set bw $args
	if { $bw != $bandwidth} {
	    Queue/PriorityQueue set limit2_ [expr int($buffers*$bw/(8*$packetSize))]
	    Queue/DropTail set limit_ [expr int($buffers*$bw/(8*$packetSize))]
	    Queue set limit_ [expr int(4*$buffers*$bw/(3*8*$packetSize))]
	    Queue/DRR set blimit_ [expr int($buffers*$bw/8)]
	} else {
	    Queue set limit_ [expr 4*$noofbuffers/3]
	    Queue/PriorityQueue set limit2_ $noofbuffers
	    Queue/DropTail set limit_ $noofbuffers
	    Queue/DRR set blimit_ [expr $noofbuffers*$packetSize]
	}
    } else {
	set bw $bandwidth
    }

    Queue/PriorityQueue set thresh_  [expr int([Queue/PriorityQueue set limit2_]*0.6)]
    Queue/RED set maxthresh_ [expr int([Queue set limit_])]
    Queue/RED set thresh_ [expr int([Queue/RED set maxthresh_]/3)]
    Queue/RED set mean_pktsize_ $packetSize
    Queue/RED set q_weight_ 0.002
    Queue/RED set linterm_ 10
    Queue/RED set drop-tail_ false
    

    #    $ns duplex-link-of-interfaces $node($a) $node($b) $bw $delay $qType
    $ns duplex-link $node($a) $node($b) $bw $delay $qType
}

Topology instproc alloc-addr {} {
	$self instvar addr_allocator
	set a $addr_allocator
	#XXX
	incr addr_allocator 32
	return $a
}

#
# build a new source (by allocating a new address) and
# place it at node $nn.  start it up at random time 
#
Topology instproc place_source { nn when level srcType } {
	#XXX
	global rates

	$self instvar node ns
	set a [$self alloc-addr]
	
	puts stderr $when
	if { $srcType == "lsm" } {
	    $ns build_lsm $level $a 4  $node($nn) $when
	} else {
	    $ns build_source_set s$a $rates 1 $a $node($nn) $when
	}

	return $a
}

##XXX not yet ported this func
Topology instproc place_my_source { nn  filename when time } {
    global qType packetSize bandwidth runtime ref
    $self instvar node ns 
    
    set a [$self alloc-addr]

    set s1 [new Agent/CBR/UDP]
    $s1 set dst_ $a
    $ns attach-agent $node($nn) $s1

    set exp1 [new Traffic/Expoo]
    $exp1 set packet-size $packetSize
    $exp1 set burst-time 100ms
    $exp1 set idle-time 100ms
    $exp1 set rate 6000e3
    $s1 attach-traffic $exp1

    set ref($a) $s1
    $ns at $when "$s1 start"
    $ns at [expr  $runtime- 0.001] "$ref($a) set npkts [$s1 set seqno_]"
    
#    set s1 [new Agent/Source]
#    $s1 set dst $a

#    if { $qType == "cdrop-tail" } {
#	$s1 set priority 1
#    }
#    $s1 set packet-size $packetSize
#    $ns attach-agent $node($nn) $s1

#    set tfile [new Tracefile]
#    $tfile file $filename

#    set trace [new Traffic/Trace]
#    $trace tracefile $tfile $time

#    $trace srcagent $s1
#    $s1 srcmdl $trace

#    $ns at $when "$s1 start"

    return  $a
}

#CBR Xtraffic sources
Topology instproc place_cbr_source { nn rate when} {
    global qType packetSize
    $self instvar node ns 
    
    set a [$self alloc-addr]

#    set s1 [new Agent/Source]
#    $s1 set dst $a

#    set s1 [new agent/cbr]
#    $s1 set cls 10
#    $s1 set packet-size 250
#    $s1 set interval_ [expr 250 * 8./32e3]
#    $s1 set interval [expr 250 * 8./32e3]
#    $s1 set random 1


    set c [new Agent/CBR]
    $c set dst_ $a
    $c set packetSize_ $packetSize
    $c set interval_ [expr $packetSize * 8./ $rate]
    $c set random_ 1


    $ns at $when "$c start"
    $ns attach-agent $node($nn) $c
    return  $a
}



Topology instproc place_my_receiver { nn addr fid when } {
    $self instvar ns
    $ns at $when "$self build_my_receiver $nn $addr $fid"
}


#MyLossTrace reqd only for Xtraffic receivers
MyLossTrace instproc get-all-layer-stats { time addr fid } {
    global plots runtime rcvTime ref

    $self instvar sumutil sumtime nlost_ npkts_ lastlost lastpkts
    
    if { $time == $rcvTime } {
	incr sumtime
	set nlost_ 0
	set npkts_ 0
    } else {
	#set frac [compute-recv-fraction $nlost_ $npkts_ $lastlost $lastpkts]
	set frac [compute-recv-fraction $npkts_ $lastpkts $fid]
	set sumutil [expr $sumutil + $frac ]
	#	puts stderr "% pkts rcvd for layer $i : $frac"
	if { $time >= $runtime  } {
	    $ref($addr) instvar npkts
	    puts $plots "$fid $addr $time $sumtime $frac [expr $sumutil*1.0/$npkts] $nlost_ $npkts_ $lastlost $lastpkts 1.0  [expr $sumutil*1.0/$npkts]  $sumtime"
	    flush $plots
	}

	incr sumtime
	set lastlost $nlost_
	set lastpkts $npkts_
    }
}

MyLossTrace instproc loss-stats { ns interval addr fid } {
    $self get-all-layer-stats [$ns now] $addr $fid
    $ns at [expr [$ns now] + $interval] "$self loss-stats $ns $interval $addr $fid"
    #puts stderr "*****At [$ns now]"

}


Topology instproc build_my_receiver { nn addr fid } {
    $self instvar node ns
    global statinterval rcvTime

#    set ctrl [$ns create-agent $node($nn) Agent/Message 0]
#    $ctrl proc handle msg "recv_ctrl $self \$msg"
#    set caddr [expr $addr +0x1000]
#    $node($nn) join-group $ctrl $caddr
#    $ctrl set dst_ $caddr


    set rcvr [new MyLossTrace]
    $ns attach-agent $node($nn) $rcvr
    $rcvr set dst_ $addr
    $node($nn) join-group $rcvr $addr
    $ns at $rcvTime "$rcvr loss-stats $ns $statinterval $addr $fid"
}


#
# build a new receiver for address $addr and
# place it at node $nn.
#
Topology instproc build_receiver { nn addr } {
	#XXX
	global nlevel statinterval rcvTime
	$self instvar node ns
	set rcvr [build_loss_monitors $ns $node($nn) $addr $nlevel]

	$ns at $rcvTime "$rcvr loss-stats $ns $statinterval $addr"

#	if { $type == "rlm" } {

#	} else {
#	    set rcvr [build_custom_receiver $ns $node($nn) $addr $nlevel]
#	}

}

Topology instproc place_receiver { nn addr when } {
	$self instvar ns
	$ns at $when "$self build_receiver $nn $addr"
}


Topology instproc place_lsm_receiver { nn addr when level } {
    $self instvar ns
    $ns at $when "$self build_lsm_receiver $nn $addr $level"
}

Topology instproc build_lsm_receiver { nn addr level } {
    global statinterval rcvTime
    $self instvar node ns
    set i 0
    puts stderr $level
    while { $i < $level } {
	puts stderr "Creating rcvr $i"
	set rcvr [ new MyLossTrace ]
	$ns attach-agent $node($nn)  $rcvr
#	$rcvr set dst_ [expr $addr ]
	$node($nn) join-group $rcvr $addr
	$ns at $rcvTime "$rcvr loss-stats $ns $statinterval $addr $i" 
	incr i
	incr addr
    }
}

Topology instproc finish {} {
}


foreach i "0 1" {
    Class Scenario$i -superclass Topology
}

Scenario0 instproc init args {
    eval $self next $args
    global bandwidth debugfile sources qType delay level srcType
    $self instvar ns node

#    set ctrmcastcomp [new CtrMcastComp $ns]
#    set ctrrpcomp [new CtrRPComp $ns]

    

#    $ns trace-all $debugfile
    $self build_link 0 1 $delay $bandwidth
#    set r [expr 60. + double([ns-random] % 10000000) / 1e7 * 60]
#    set addr [$self place_source 0 $r]
#    $self place_receiver 1 $addr $r 

    set i 2    

    while { $i <= [expr $sources + 1] } {
	$self build_link 0 $i 0.01 10e6

	set r [expr 60. + double([ns-random] % 10000000) / 1e7 * 60]
	set addr [$self place_source $i $r $level $srcType]
	puts stderr "Placing a source with $addr"
	$self place_receiver 1 $addr $r 
	incr i
    }

#    set mproto CtrMcast
#    set mrthandle [$ns $mproto {} ]


#    set ctrmcast0 [new CtrMcast $ns $node(0) $ctrmcastcomp [list 1 1]]
#    set ctrmcast1 [new CtrMcast $ns $node(1) $ctrmcastcomp [list 0 1]]
#    set ctrmcast2 [new CtrMcast $ns $node(2) $ctrmcastcomp [list 0 1]]
#    set ctrmcast3 [new CtrMcast $ns $node(3) $ctrmcastcomp [list 0 1]]
#    set ctrmcast4 [new CtrMcast $ns $node(4) $ctrmcastcomp [list 0 1]]
#    set ctrmcast5 [new CtrMcast $ns $node(5) $ctrmcastcomp [list 0 1]]
#    set ctrmcast6 [new CtrMcast $ns $node(6) $ctrmcastcomp [list 0 1]]
#    set ctrmcast7 [new CtrMcast $ns $node(7) $ctrmcastcomp [list 0 1]]
#    set ctrmcast8 [new CtrMcast $ns $node(8) $ctrmcastcomp [list 0 1]]
#    set ctrmcast9 [new CtrMcast $ns $node(9) $ctrmcastcomp [list 0 1]]
#    $ns compute-routes
#    $ctrrpcomp compute-rpset

    DM set PruneTimeout 1000
    set mproto DM
    set mrthandle [$ns mrtproto $mproto {} ]

#    set i 0
#    set addr [expr 0x8000]
#    while { $i < $sources} {
#	set j 0
#	while { $j < $level } {
#	    puts stderr "Switching treetype for [expr $addr+$j]"
#	    $ns at 0.0 "$ctrmcastcomp switch-treetype [expr $addr+$j]"
#	    incr j
#	}
#	incr addr 32
#	incr i
#    }
	
    if [info exists debugfile] { 
	$ns trace-queue $node(2) $node(0) $debugfile
#	$ns trace-queue $node(0) $node(1) $debugfile
    }


}



Scenario1 instproc init args {
    eval $self next $args
    global M bandwidth debugfile sources qType srcType bgsrcType delay bgrates burstylength bglevel level traffic

    $self instvar ns node 

#    set ctrmcastcomp [new CtrMcastComp $ns]
#    set ctrrpcomp [new CtrRPComp $ns]


#    $ns trace-all $debugfile

    $self build_link 0 1 $delay $bandwidth
    $self build_link 0 2 0.01 10e6


    set r [expr 60. + double([ns-random] % 10000000) / 1e7 * 60]
    set addr [$self place_source 2 $r $level $srcType]
    puts stderr "Src address is $addr"
    $self place_receiver 1 $addr $r 

#    set traffic 2
    set i 3
    while { $i <= [expr $sources + 1] } {
	set r [expr 60. + double([ns-random] % 10000000) / 1e7 * 60]
	$self build_link 0 $i 0.01 [expr 10 * $bandwidth]

	set layerno 0
	#set time [expr [ns-random] % 100000 /1e5 * $burstylength ]
        set time 0	

	if { $bgsrcType != "lsm" } {
	    foreach level $bgrates {
		if { $bgsrcType == "cbr" } {
		    set addr [$self place_cbr_source $i $level $r]
		} else {
		    if { $bgsrcType == "bursty" } {
			set addr [$self place_my_source $i $level $r $time]
			#	    set addr [$self place_my_source $i 0 $r $time]
		    } 
		}
		
		puts stderr $r
		$self place_my_receiver 1 $addr $layerno $r
		incr layerno
	    }
	} else {
	    # fix this to fit in background src model traffic
	    #set r [expr 60. + double([ns-random] % 10000000) / 1e7 * 60]
	    set addr [$self place_source $i $r $bglevel $bgsrcType]
	    $self place_lsm_receiver 1 $addr $r $bglevel
	}
	incr i
    }

#    set ctrmcast0 [new CtrMcast $ns $node(0) $ctrmcastcomp [list 1 1]]
#    set ctrmcast1 [new CtrMcast $ns $node(1) $ctrmcastcomp [list 0 1]]
#    set ctrmcast2 [new CtrMcast $ns $node(2) $ctrmcastcomp [list 0 1]]
    DM set PruneTimeout 1000
    set mproto DM
    set mrthandle [$ns mrtproto $mproto {} ]
#    set i 3
#    while { $i < [expr $sources + 2 ] } {
#	set ctrmcast [new CtrMcast $ns $node($i) $ctrmcastcomp [list 0 1]]
	incr i
#    }
#    $ns compute-routes
#    $ctrrpcomp compute-rpset

#    set i 0
#    set addr [expr 0x8000 ]
#    set j 0
#    while { $j < $level } {
#	puts stderr "Switching treetype for [expr $addr+$j]"
#	$ns at 0.0 "$ctrmcastcomp switch-treetype [expr $addr+$j]"
#	incr j
#    }
#    incr addr 32
#    while { $i < [expr $sources -1] } {
#	set j 0
#	while { $j < $bglevel } {
#	    puts stderr "Switching treetype for [expr $addr+$j]"
#	    $ns at 0.0 "$ctrmcastcomp switch-treetype [expr $addr+$j]"
#	    incr j
#	}
#	incr addr 32
#	incr i
#    }

    if [info exists debugfile] { 
	$ns trace-queue $node(0) $node(1) $debugfile
    }
}

#XXX
if [info exists rlmTraceFile] {
	set rlmTraceFile [open $rlmTraceFile w]
}

#
# set up a constant bit-rate stream at $node with traffic class $class
# with destination $addr and rate $bw (in bits per second)
#
Simulator instproc cbr_flow { node class addr bw } {
    global packetSize
    set agent [$self create-agent $node Agent/CBR $class]
    #	$agent node $node
    #XXX abstraction violation
    $agent set dst_ $addr
    
    #	$agent connect $addr 0
    #	$agent set class $class
    $agent set packetSize_ $packetSize
    #XXX set both interval and interval_ because we have two
    #versions of ns floating around that have either
    $agent set interval_ [expr $packetSize * 8. / $bw]
    $agent set interval [expr $packetSize * 8. / $bw]
    $agent set random_ 1
    return $agent
}


#source ../qos/traffic-trace.tcl
source layers.tcl
# create a layered flow at $node.  Start it up at time $when.
# use the list of rates in the list $rates, and use $baseAddr
# as the base address of the consecutive multicast group (MMG).
# $baseClass is the traffic class for the base flow.
#
Simulator instproc build_source_set { mmgName rates baseClass baseAddr node when } {
	global src_mmg src_rate srcType burstylength
	set i 0
	foreach r $rates {
	    set src_rate([expr $baseAddr]) $r
	    set k $mmgName:$i
	    incr i
	    if { $srcType == "cbr" } {
		set src_mmg($k) [$self cbr_flow $node $baseClass $baseAddr $r]
	    } 
	    if { $srcType == "bursty" } {
		set time [expr [ns-random] % 100000 / 1e5 * $burstylength ]
		set src_mmg($k) [$self fd_flow $node $baseClass $baseAddr $r $time]
	    }
	    $self at $when "$src_mmg($k) start"
	    incr baseClass
	    incr baseAddr
	}
}

Simulator instproc build_lsm { layers baseAddr baserate node when } {
    global packetSize lsminterval upinterval runtime rcvTime
    set src [new LayeredSource $baseAddr $layers $packetSize $lsminterval $baserate $self $upinterval 0 ]
    $self attach-layers $node  $src
    $self at $when "$src start"
    $self at [expr $runtime-0.001] "output-pkts $src $layers"
    $self at $rcvTime "zero-pkts $src $layers"
}

proc zero-pkts { src layers } {
    $src instvar layer
    set i 1
    while { $i <= $layers } {
	$layer($i) set npkts 0
	incr i
    }
}

proc output-pkts { src layers} { 
    global totalpkts
    $src instvar layer
    set i 1
    while { $i <= $layers } {
	$layer($i) instvar dst_ npkts
	puts stderr "$dst_ : $npkts"
	#puts $totalpkts "$dst_ : $npkts"
	incr i
    }
}

Simulator instproc finish {} {
#XXX
	global nlevel rcvrMMG graph fig dir rectFile scenario lossTraceFile debugfile qType srcType bgsrcType outfilename  delay runno noofbuffers totalpkts
	puts finish
	#XXX

	flush_all_trace

	if [info exists plots] {
		close $plots
	}


	close $totalpkts

	#XXX
	rlm_finish

#	if { $proto == "rlm" } { 
#	    $proto\_finish
#	}
	#XXX
	if [info exists lossTraceFile] {
		close $lossTraceFile
	}

	if [info exists debugfile] {
		close $debugfile
	}


#	if { $proto == "rlm" } {
#		if $graph {
#			exec ./rect_graph $rectFile &
#		}
#		if $fig {
#			exec ./rect_fig $rectFile &
#		}
#	}
	#XXX
#	$self instvar topology
#	$topology finish
#	$self finish_animator nam-$proto-$scenario

	if [info exists rcvrMMG] {
	#	puts [expr 8192.0 * [node_loss $rcvrMMG] / [total_bits $rcvrMMG]]
		set br [expr [total_bits $rcvrMMG] / 8.]
		set bd [total_bytes_delivered $rcvrMMG]
	#	puts $br/$bd/[expr $br / $bd]
		puts "loss-frac [expr 1024. * [node_loss $rcvrMMG] / $bd] \
			goodput [expr $bd / [optimal_bytes]]"
	}

	if { $bgsrcType == "null" } {
	   exec awkfile0 $outfilename  $dir/$qType-rlm-$srcType-$bgsrcType-$nlevel-$delay-$noofbuffers-$runno
	} else {
	    exec awkfile.rlm $outfilename $dir/$qType-rlm-$srcType-$bgsrcType-$nlevel-$delay-$noofbuffers-$runno
	}
	exit 0
}

Simulator instproc tick {} {
	puts stderr [$self now]
	$self at [expr [$self now] + 30.] "$self tick"
}

#Simulator instproc run { top runtime } {
#    $self instvar topology
#    set topology $top
#    $self at [expr $runtime + 1] "$self finish"
#    $self next
#}


#gets loss stats from all layers at interval
MMG instproc get-all-layer-stats { time addr} {
    global plots runtime rcvTime ref
    $self instvar layers
    set i 0

#    if {! [ info exists utilfunc ] } {
#	set utilfunc trivial-loss-util
#    }


    foreach s $layers {
	$s instvar util
	set monitor [$s mon]
	$monitor instvar sumutil sumtime  lastlost lastpkts mynlost mynpkts nlost_ npkts_ init timein

	if  { $time == $rcvTime } {
		incr sumtime
		set timein 0
		set mynlost $nlost_
		set mynpkts $npkts_
	} else {
		if { $i <= [expr [$self subscription] -1] } {
			set frac [compute-recv-fraction [expr $npkts_-$mynpkts] $lastpkts $i]
			incr timein
		} else {
			set frac 0
		}
	
	    

#	    if { $lag == 1 } { 
#		set frac [$util $utilfunc [$self compute-min-lag [compute-recv-fraction [expr $nlost_-$mynlost] [expr $npkts_-$mynpkts] $lastlost $lastpkts ]] ] 
#	    } else {
#		if { $i <= [expr [$self subscription] -1] } {
		    #		set frac [$util $utilfunc [compute-recv-fraction [expr $nlost_-$mynlost] [expr $npkts_-$mynpkts] $lastlost $lastpkts ] ] 
		    #set frac [$util $utilfunc [compute-recv-fraction [expr $nlost_-$mynlost] $lastlost $i] ]
#		    set frac [$util $utilfunc [compute-recv-fraction [expr $npkts_-$mynpkts] $lastpkts $i ]]
#		    incr timein
#		} else {
#		    set frac 0
#		}
#	    }
	    

		
		set sumutil [expr $sumutil + $frac ]
		#	puts stderr "% pkts rcvd for layer $i : $frac"
		if { $time >= $runtime } {
			$ref($addr) instvar npkts
			if { [expr $timein*4.0*($i+1)] > $npkts } {
				set ngen $npkts
			} else {
				set ngen [expr $timein*4.0*($i+1)]
			}
			if { $ngen != 0 } {  
				puts $plots "$i $addr $time $sumtime $frac [expr $sumutil*1.0/$npkts] [expr $nlost_-$mynlost] [expr $npkts_-$mynpkts] $lastlost $lastpkts [expr $timein*1.0/$sumtime] [expr $sumutil*1.0/$ngen] $timein"
			} else {
				puts $plots "$i $addr $time $sumtime $frac [expr $sumutil*1.0/$npkts] [expr $nlost_-$mynlost] [expr $npkts_-$mynpkts] $lastlost $lastpkts 0 0 $timein"
			}
			flush $plots
		}
		incr sumtime
		set lastlost [expr $nlost_- $mynlost]
		set lastpkts [expr $npkts_ - $mynpkts]
		incr i 
		incr addr
	}
    }
}

MMG instproc loss-stats { ns interval addr } {


    $self get-all-layer-stats [$ns now] $addr

    $ns at [expr [$ns now] + $interval] "$self loss-stats $ns $interval $addr"
 #   puts stderr "*****At [$ns now]"

}


MyLossTrace instproc compute-recv-fraction {} {
    global lag
    $self instvar mynlost mynpkts lastlost lastpkts sumutil sumtime
    #puts stderr "Diff Recv Pkts : [expr $npkts - $lastpkts] Diff Lost Pkts : [expr $mynlost - $lastlost]" 
    if { $mynpkts != 0 }  {
	if { $mynpkts != $lastpkts } {
	    set frac [expr ($mynpkts - $lastpkts) * 1.0 / ($mynlost + $mynpkts - $lastlost - $lastpkts)]
	} else {
	    set frac 0
	}
    } else {
	set frac 0
    }
    

    if { $lag == 1 } { 
	return [$self compute-min-lag $frac]
    } else {
	return $frac
    }
}

#Each MMG has a Losstrace corresponding to every level
#proc compute-recv-fraction { nlost npkts lastlost lastpkts } {
#    if { $npkts != 0 }  {
#	if { $npkts != $lastpkts } {
#	    set frac [expr ($npkts - $lastpkts) * 1.0 / ($nlost + $npkts - $lastlost - $lastpkts)]
#	} else {
#	    set frac 0
#	}
#    } else {
#	set frac 0
#   }
    
#    return $frac
#}

#proc compute-recv-fraction { nlost lastlost layerno } {
#    return [expr 1-($nlost-$lastlost)/(4*(pow(2,($layerno%32))))]
#}

proc compute-recv-fraction { npkts lastpkts layer } {
    return [expr $npkts-$lastpkts]
}

set ns [new Simulator]
Simulator set EnableMcast_ 1
Simulator set NumberInterfaces_ 1


set scn [new Scenario$scenario $ns]
$ns at [expr $runtime +1] "$ns finish"
$ns run
